/*
 * Copyright (c) "Neo4j"
 * Neo4j Sweden AB [http://neo4j.com]
 *
 * This file is part of Neo4j.
 *
 * Neo4j is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.neo4j.cypher.internal.compiler.ast.convert.plannerQuery

import org.neo4j.cypher.internal.compiler.planner.LogicalPlanningTestSupport
import org.neo4j.cypher.internal.expressions.NodePattern
import org.neo4j.cypher.internal.expressions.PatternExpression
import org.neo4j.cypher.internal.expressions.RelTypeName
import org.neo4j.cypher.internal.expressions.RelationshipChain
import org.neo4j.cypher.internal.expressions.RelationshipPattern
import org.neo4j.cypher.internal.expressions.RelationshipsPattern
import org.neo4j.cypher.internal.expressions.SemanticDirection
import org.neo4j.cypher.internal.ir.PatternRelationship
import org.neo4j.cypher.internal.ir.Predicate
import org.neo4j.cypher.internal.ir.Selections
import org.neo4j.cypher.internal.ir.SimplePatternLength
import org.neo4j.cypher.internal.ir.helpers.ExpressionConverters.asQueryGraph
import org.neo4j.cypher.internal.util.AnonymousVariableNameGenerator
import org.neo4j.cypher.internal.util.test_helpers.CypherFunSuite

class PatternExpressionConverterTest extends CypherFunSuite with LogicalPlanningTestSupport {

  private val aNode = NodePattern(Some(varFor("a")), Seq.empty, None, None)_
  private val bNode = NodePattern(Some(varFor("b")), Seq.empty, None, None)_
  private val rRel = RelationshipPattern(Some(varFor("r")), Seq.empty, None, None, SemanticDirection.OUTGOING)_
  private val TYP: RelTypeName = RelTypeName("TYP")_
  private val dependencies = Set("a", "r", "b")

  private val rRelWithType = rRel.copy(types = Seq(TYP)) _
  private val planRel = PatternRelationship("r", ("a", "b"), SemanticDirection.OUTGOING, Seq.empty, SimplePatternLength)
  private val planRelWithType = PatternRelationship("r", ("a", "b"), SemanticDirection.OUTGOING, Seq(TYP), SimplePatternLength)

  test("(a)-[r]->(b)") {
    // Given
    val patternExpression = createPatternExpression(aNode, rRel, bNode)

    // When
    val qg = asQueryGraph(patternExpression, dependencies, new AnonymousVariableNameGenerator)

    // Then
    qg.selections should equal(Selections())
    qg.patternRelationships should equal(Set(planRel))
    qg.argumentIds should equal(dependencies)
    qg.patternNodes should equal(Set("a", "b"))
  }

  test("(a)-[r:TYP]->(b)") {
    // Given
    val patternExpression = createPatternExpression(aNode, rRelWithType, bNode)

    // When
    val qg = asQueryGraph(patternExpression, dependencies, new AnonymousVariableNameGenerator)

    // Then
    qg.selections should equal(Selections())
    qg.patternRelationships should equal(Set(planRelWithType))
    qg.argumentIds should equal(dependencies)
    qg.patternNodes should equal(Set("a", "b"))
  }

  test("(a)-[r]->(b:Label)") {
    // Given
    val patternExpression = createPatternExpression(aNode, rRel, bNode.copy(labels = Seq(labelName("Label")))(pos))

    // When
    val qg = asQueryGraph(patternExpression, dependencies, new AnonymousVariableNameGenerator)

    // Then
    val predicate = hasLabels("b", "Label")
    qg.selections should equal(Selections(Set(Predicate(Set("b"), predicate))))
    qg.patternRelationships should equal(Set(planRel))
    qg.argumentIds should equal(dependencies)
    qg.patternNodes should equal(Set("a", "b"))
  }

  def createPatternExpression(n1: NodePattern, r: RelationshipPattern, n2: NodePattern): PatternExpression =
    PatternExpression(RelationshipsPattern(RelationshipChain(n1, r, n2) _) _)(dependencies.map(varFor), "", "")
}
